
import ctypes
import importlib.util

from numpy.ctypeslib import ndpointer

# control structures
class glp_smcp(ctypes.Structure):
    _fields_ = [
        ('msg_lev', ctypes.c_int),
        ('meth', ctypes.c_int),
        ('pricing', ctypes.c_int),
        ('r_test', ctypes.c_int),
        ('tol_bnd', ctypes.c_double),
        ('tol_dj', ctypes.c_double),
        ('tol_piv', ctypes.c_double),
        ('obj_ll', ctypes.c_double),
        ('obj_ul', ctypes.c_double),
        ('it_lim', ctypes.c_int),
        ('tm_lim', ctypes.c_int),
        ('out_frq', ctypes.c_int),
        ('out_dly', ctypes.c_int),
        ('presolve', ctypes.c_int),
        ('excl', ctypes.c_int),
        ('shift', ctypes.c_int),
        ('aorn', ctypes.c_int),
        ('foobar', ctypes.c_double*33),
    ]

class glp_iptcp(ctypes.Structure):
    _fields_ = [
        ('msg_lev', ctypes.c_int),
        ('ord_alg', ctypes.c_int),
        ('foo_bar', ctypes.c_double*48),
    ]

class glp_mpscp(ctypes.Structure):
    _fields_ = []

class glp_bfcp(ctypes.Structure):
    _fields_ = [
        ('msg_lev', ctypes.c_int),       # (not used)
        ('type', ctypes.c_int),          # factorization type
        ('lu_size', ctypes.c_int),       # (not used)
        ('piv_tol', ctypes.c_double),    # sgf_piv_tol
        ('piv_lim', ctypes.c_int),       # sgf_piv_lim
        ('suhl', ctypes.c_int),          # sgf_suhl
        ('eps_tol', ctypes.c_double),    # sgf_eps_tol
        ('max_gro', ctypes.c_double),    # (not used)
        ('nfs_max', ctypes.c_int),       # fhvint.nfs_max
        ('upd_tol', ctypes.c_double),    # (not used)
        ('nrs_max', ctypes.c_int),       # scfint.nn_max
        ('rs_size', ctypes.c_int),       # (not used)
        ('foo_bar', ctypes.c_double*38), # (reserved)
    ]

# integer optimizer control parameters
class glp_tree(ctypes.Structure):
    _fields_ = []
class glp_iocp(ctypes.Structure):
    _fields_ = [
        ('msg_lev', ctypes.c_int),       # message level (see glp_smcp)
        ('br_tech', ctypes.c_int),       # branching technique
        ('bt_tech', ctypes.c_int),       # backtracking technique:
        ('tol_int', ctypes.c_double),    # mip.tol_int
        ('tol_obj', ctypes.c_double),    # mip.tol_obj
        ('tm_lim', ctypes.c_int),        # mip.tm_lim (milliseconds)
        ('out_frq', ctypes.c_int),       # mip.out_frq (milliseconds)
        ('out_dly', ctypes.c_int),       # mip.out_dly (milliseconds)
        #void (*cb_func)(glp_tree *T, void *info);  # mip.cb_func
        ('cb_func', ctypes.CFUNCTYPE(None, ctypes.POINTER(glp_tree), ctypes.c_void_p)),
        ('cb_info', ctypes.c_void_p),    # mip.cb_info
        ('cb_size', ctypes.c_int),       # mip.cb_size
        ('pp_tech', ctypes.c_int),       # preprocessing technique:
        ('mip_gap', ctypes.c_double),    # relative MIP gap tolerance
        ('mir_cuts', ctypes.c_int),      # MIR cuts       (GLP_ON/GLP_OFF)
        ('gmi_cuts', ctypes.c_int),      # Gomory's cuts  (GLP_ON/GLP_OFF)
        ('cov_cuts', ctypes.c_int),      # cover cuts     (GLP_ON/GLP_OFF)
        ('clq_cuts', ctypes.c_int),      # clique cuts    (GLP_ON/GLP_OFF)
        ('presolve', ctypes.c_int),      # enable/disable using MIP presolver
        ('binarize', ctypes.c_int),      # try to binarize integer variables
        ('fp_heur', ctypes.c_int),       # feasibility pump heuristic
        ('ps_heur', ctypes.c_int),       # proximity search heuristic
        ('ps_tm_lim', ctypes.c_int),     # proxy time limit, milliseconds
        ('sr_heur', ctypes.c_int),       # simple rounding heuristic
        ('use_sol', ctypes.c_int),       # use existing solution
        ('save_sol', ctypes.c_char_p),   # filename to save every new solution
        ('alien', ctypes.c_int),         # use alien solver
        ('flip', ctypes.c_int),          # use long-step dual simplex; not documented--should not be used
        ('foo_bar', ctypes.c_double*23), # (reserved)
    ]


# LP problem structure
class glp_prob(ctypes.Structure):
    class DMP(ctypes.Structure):
        _fields_ = []
    class glp_tree(ctypes.Structure):
        _fields_ = []
    class GLPROW(ctypes.Structure):
        _fields_ = []
    class GLPCOL(ctypes.Structure):
        _fields_ = []
    class AVL(ctypes.Structure):
        _fields_ = []
    class BFD(ctypes.Structure):
        _fields_ = []
    # shouldn't access these directly!
    _fields_ = [
        ('pool', ctypes.POINTER(DMP)),
        ('tree', ctypes.POINTER(glp_tree)),
        ('name', ctypes.c_char_p),
        ('obj', ctypes.c_char_p),
        ('dir', ctypes.c_int),
        ('c0', ctypes.c_double),
        ('m_max', ctypes.c_int),
        ('n_max', ctypes.c_int),
        ('m', ctypes.c_int),
        ('n', ctypes.c_int),
        ('nnz', ctypes.c_int),
        ('row', ctypes.POINTER(ctypes.POINTER(GLPROW))),
        ('col', ctypes.POINTER(ctypes.POINTER(GLPCOL))),
        ('r_tree', ctypes.POINTER(AVL)),
        ('c_tree', ctypes.POINTER(AVL)),
        ('valid', ctypes.c_int),
        ('head', ctypes.POINTER(ctypes.c_int)),
        ('bfd', ctypes.POINTER(BFD)),
        ('pbs_stat', ctypes.c_int),
        ('dbs_stat', ctypes.c_int),
        ('obj_val', ctypes.c_double),
        ('it_cnt', ctypes.c_int),
        ('some', ctypes.c_int),
        ('ipt_stat', ctypes.c_int),
        ('ipt_obj', ctypes.c_double),
        ('mip_stat', ctypes.c_int),
        ('mip_obj', ctypes.c_double),
    ]

class GLPK:

    # Where is GLPK? Find where the C extension module was installed:
    _glpk_lib_path = importlib.util.find_spec("glpk4_65").origin


    INT_MAX = ctypes.c_uint(-1).value // 2
    GLP_ON = 1
    GLP_OFF = 0

    # solution status:
    GLP_UNDEF = 1   # solution is undefined
    GLP_FEAS = 2    # solution is feasible
    GLP_INFEAS = 3  # solution is infeasible
    GLP_NOFEAS = 4  # no feasible solution exists
    GLP_OPT = 5     # solution is optimal
    GLP_UNBND = 6   # solution is unbounded
    STATUS_CODES = {
        GLP_UNDEF: 'solution is undefined',
        GLP_FEAS: 'solution is feasible',
        GLP_INFEAS: 'solution is infeasible',
        GLP_NOFEAS: 'no feasible solution exists',
        GLP_OPT: 'solution is optimal',
        GLP_UNBND: 'solution is unbounded',
    }

    # Pricing techniques
    GLP_PT_STD = 17 # 0x11 # standard (Dantzig's rule)
    GLP_PT_PSE = 34 # 0x22 # projected steepest edge

    # Ratio test techniques
    GLP_RT_STD = 17 # 0x11
    GLP_RT_HAR = 34 # 0x22
    GLP_RT_FLIP = 51 # 0x33

    # Objective sense
    GLP_MIN = 1
    GLP_MAX = 2

    # Bound types
    GLP_FR = 1 # -inf <  x  < +inf
    GLP_LO = 2 #   lb <= x  < +inf
    GLP_UP = 3 # -inf <  x <= ub
    GLP_DB = 4 #   lb <= x <= ub
    GLP_FX = 5 #   lb == x == ub

    # Scaling techniques
    GLP_SF_GM = 1     # 0x01   # geometric scaling
    GLP_SF_EQ = 16    # 0x10   # equilibration scaling
    GLP_SF_2N = 3     # 0x20   # round scale factors to nearest power of two
    GLP_SF_SKIP = 64  # 0x40 # skip scaling, if the problem is well scaled
    GLP_SF_AUTO = 128 # 0x80 # choose scaling options automatically

    # return codes for glp_simplex driver
    SUCCESS = 0
    GLP_EBADB = 1
    GLP_ESING = 2
    GLP_ECOND = 3
    GLP_EBOUND = 4
    GLP_EFAIL = 5
    GLP_EOBJLL = 6
    GLP_EOBJUL = 7
    GLP_EITLIM = 8
    GLP_ETMLIM = 9
    GLP_ENOPFS = 10
    GLP_ENODFS = 11
    GLP_EROOT = 12
    GLP_ESTOP = 13
    GLP_EMIPGAP = 14
    GLP_ENOFEAS = 15
    GLP_ENOCVG = 16
    GLP_EINSTAB = 17
    GLP_EDATA = 18
    GLP_ERANGE = 19

    RET_CODES = {
        SUCCESS: 'LP problem instance has been successfully solved',
        GLP_EBADB:'invalid basis',
        GLP_ESING:'singular matrix',
        GLP_ECOND:'ill-conditioned matrix',
        GLP_EBOUND:'invalid bounds',
        GLP_EFAIL:'solver failed',
        GLP_EOBJLL:'objective lower limit reached',
        GLP_EOBJUL:'objective upper limit reached',
        GLP_EITLIM:'iteration limit exceeded',
        GLP_ETMLIM:'time limit exceeded',
        GLP_ENOPFS: 'no primal feasible solution',
        GLP_ENODFS: 'no dual feasible solution',
        GLP_EROOT: 'root LP optimum not provided',
        GLP_ESTOP: 'search terminated by application',
        GLP_EMIPGAP: 'relative mip gap tolerance reached',
        GLP_ENOFEAS: 'no primal/dual feasible solution',
        GLP_ENOCVG: 'no convergence',
        GLP_EINSTAB: 'numerical instability',
        GLP_EDATA: 'invalid data',
        GLP_ERANGE: 'result out of range',
    }

    # Message levels
    GLP_MSG_OFF = 0
    GLP_MSG_ERR = 1
    GLP_MSG_ON = 2
    GLP_MSG_ALL = 3
    GLP_MSG_DBG = 4

    # Simplex methods
    GLP_PRIMAL = 1
    GLP_DUALP = 2
    GLP_DUAL = 3

    # Ordering strategies
    GLP_ORD_NONE = 0
    GLP_ORD_QMD = 1
    GLP_ORD_AMD = 2
    GLP_ORD_SYMAMD = 3

    # MPS formats
    GLP_MPS_DECK = 1  # fixed (ancient)
    GLP_MPS_FILE = 2  # free (modern)

    # Factorization strategies
    GLP_BF_LUF = 0  # 0x00  # plain LU-factorization
    GLP_BF_BTF = 16 # 0x10  # block triangular LU-factorization
    GLP_BF_FT = 1   # 0x01  # Forrest-Tomlin (LUF only)
    GLP_BF_BG = 2   # 0x02  # Schur compl. + Bartels-Golub
    GLP_BF_GR = 3   # 0x03  # Schur compl. + Givens rotation

    # Branching techniques
    GLP_BR_FFV = 1  # first fractional variable
    GLP_BR_LFV = 2  # last fractional variable
    GLP_BR_MFV = 3  # most fractional variable
    GLP_BR_DTH = 4  # heuristic by Driebeck and Tomlin
    GLP_BR_PCH = 5  # hybrid pseudocost heuristic

    # Backtracking technique
    GLP_BT_DFS = 1  # depth first search
    GLP_BT_BFS = 2  # breadth first search
    GLP_BT_BLB = 3  # best local bound
    GLP_BT_BPH = 4  # best projection heuristic

    # MIP preprocessing
    GLP_PP_NONE =  0 # disable preprocessing
    GLP_PP_ROOT =  1 # preprocessing only on root level
    GLP_PP_ALL = 2   # preprocessing on all levels

    # Variable types
    GLP_CV = 1 # continuous variable
    GLP_IV = 2 # integer variable
    GLP_BV = 3 # binary variable

    def __init__(self):

        # Load the shared library;
        _lib = ctypes.cdll.LoadLibrary(self._glpk_lib_path)

        _lib.glp_version.restype = ctypes.c_char_p

        _lib.glp_create_prob.restype = ctypes.POINTER(glp_prob)

        # Setters
        _lib.glp_set_prob_name.restype = None
        _lib.glp_set_prob_name.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_char_p]

        _lib.glp_set_obj_name.restype = None
        _lib.glp_set_obj_name.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_char_p]

        _lib.glp_set_obj_dir.restype = None
        _lib.glp_set_obj_dir.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_add_cols.restype = ctypes.c_int
        _lib.glp_add_cols.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_set_obj_coef.restype = None
        _lib.glp_set_obj_coef.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int, ctypes.c_double]

        _lib.glp_set_col_name.restype = None
        _lib.glp_set_col_name.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int, ctypes.c_char_p]

        _lib.glp_set_col_bnds.restype = None
        _lib.glp_set_col_bnds.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int,    # col index (1-based)
            ctypes.c_int,    # type
            ctypes.c_double, # lb
            ctypes.c_double, # up
        ]

        _lib.glp_add_rows.restype = ctypes.c_int
        _lib.glp_add_rows.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_load_matrix.restype = None
        _lib.glp_load_matrix.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int,                                     # nnz
            ndpointer(ctypes.c_int, flags='C_CONTIGUOUS'),    # row indices
            ndpointer(ctypes.c_int, flags='C_CONTIGUOUS'),    # col indices
            ndpointer(ctypes.c_double, flags='C_CONTIGUOUS'), # values
        ]

        _lib.glp_set_row_bnds.restype = None
        _lib.glp_set_row_bnds.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int,    # row index (1-based)
            ctypes.c_int,    # type
            ctypes.c_double, # lb
            ctypes.c_double, # ub
        ]

        _lib.glp_scale_prob.restype = None
        _lib.glp_scale_prob.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_std_basis.restype = None
        _lib.glp_std_basis.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_adv_basis.restype = None
        _lib.glp_adv_basis.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_cpx_basis.restype = None
        _lib.glp_cpx_basis.argtypes = [ctypes.POINTER(glp_prob)]

        # Getters
        _lib.glp_get_prob_name.restype = ctypes.c_char_p
        _lib.glp_get_prob_name.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_obj_name.restype = ctypes.c_char_p
        _lib.glp_get_obj_name.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_obj_dir.restype = ctypes.c_int
        _lib.glp_get_obj_dir.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_obj_coef.restype = ctypes.c_double
        _lib.glp_get_obj_coef.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_get_num_cols.restype = ctypes.c_int
        _lib.glp_get_num_cols.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_col_name.restype = ctypes.c_char_p
        _lib.glp_get_col_name.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_get_row_name.restype = ctypes.c_char_p
        _lib.glp_get_row_name.argtypes = [ctypes.POINTER(glp_prob), ctypes.c_int]

        _lib.glp_get_num_rows.restype = ctypes.c_int
        _lib.glp_get_num_rows.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_num_nz.restype = ctypes.c_int
        _lib.glp_get_num_nz.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_mat_row.restype = ctypes.c_int # len of col indices and values
        _lib.glp_get_mat_row.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # ith row
            ctypes.POINTER(ctypes.c_int), # col indices
            ctypes.POINTER(ctypes.c_double), # values
        ]

        _lib.glp_get_row_type.restype = ctypes.c_int
        _lib.glp_get_row_type.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # ith row
        ]

        _lib.glp_get_row_lb.restype = ctypes.c_double
        _lib.glp_get_row_lb.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # ith row
        ]

        _lib.glp_get_row_ub.restype = ctypes.c_double
        _lib.glp_get_row_ub.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # ith row
        ]

        _lib.glp_get_col_type.restype = ctypes.c_int
        _lib.glp_get_col_type.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # jth col
        ]

        _lib.glp_get_col_lb.restype = ctypes.c_double
        _lib.glp_get_col_lb.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # jth col
        ]

        _lib.glp_get_col_ub.restype = ctypes.c_double
        _lib.glp_get_col_ub.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # jth col
        ]

        _lib.glp_get_status.restype = ctypes.c_int
        _lib.glp_get_status.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_obj_val.restype = ctypes.c_double
        _lib.glp_get_obj_val.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_get_col_prim.restype = ctypes.c_double
        _lib.glp_get_col_prim.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # primal value of jth col
        ]

        _lib.glp_get_col_dual.restype = ctypes.c_double
        _lib.glp_get_col_dual.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # dual value of jth col
        ]

        _lib.glp_get_row_dual.restype = ctypes.c_double
        _lib.glp_get_row_dual.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # dual value of ith row
        ]
        
        # Interior point variants
        _lib.glp_ipt_status.restype = ctypes.c_int
        _lib.glp_ipt_status.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_ipt_obj_val.restype = ctypes.c_double
        _lib.glp_ipt_obj_val.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_ipt_col_prim.restype = ctypes.c_double
        _lib.glp_ipt_col_prim.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # primal value of jth col
        ]

        _lib.glp_ipt_col_dual.restype = ctypes.c_double
        _lib.glp_ipt_col_dual.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # dual value of jth col
        ]

        # MIP variants
        _lib.glp_mip_status.restype = ctypes.c_int
        _lib.glp_mip_status.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_mip_obj_val.restype = ctypes.c_double
        _lib.glp_mip_obj_val.argtypes = [ctypes.POINTER(glp_prob)]

        _lib.glp_mip_col_val.restype = ctypes.c_double
        _lib.glp_mip_col_val.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # jth col
        ]

        _lib.glp_set_col_kind.restype = None
        _lib.glp_set_col_kind.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int, # jth col
            ctypes.c_int, # type
        ]

        # Initialize control structures
        _lib.glp_init_smcp.restype = ctypes.c_int
        _lib.glp_init_smcp.argtypes = [ctypes.POINTER(glp_smcp)]

        _lib.glp_init_iptcp.restype = ctypes.c_int
        _lib.glp_init_iptcp.argtypes = [ctypes.POINTER(glp_iptcp)]

        _lib.glp_init_iocp.restype = None
        _lib.glp_init_iocp.argtypes = [ctypes.POINTER(glp_iocp)]

        # Simplex drivers
        _lib.glp_simplex.restype = ctypes.c_int
        _lib.glp_simplex.argtypes = [ctypes.POINTER(glp_prob), ctypes.POINTER(glp_smcp)]

        _lib.glp_exact.restype = ctypes.c_int
        _lib.glp_exact.argtypes = [ctypes.POINTER(glp_prob), ctypes.POINTER(glp_smcp)]

        # Interior point drivers
        _lib.glp_interior.restype = ctypes.c_int
        _lib.glp_interior.argtypes = [ctypes.POINTER(glp_prob), ctypes.POINTER(glp_iptcp)]

        # MIP drivers
        _lib.glp_intopt.restype = ctypes.c_int
        _lib.glp_intopt.argtypes = [ctypes.POINTER(glp_prob), ctypes.POINTER(glp_iocp)]

        # Cleanup
        _lib.glp_delete_prob.restype = None
        _lib.glp_delete_prob.argtypes = [ctypes.POINTER(glp_prob)]

        # Utility
        _lib.glp_read_mps.restype = ctypes.c_int
        _lib.glp_read_mps.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int,              # format (GLP_MPS_DECK or GLP_MPS_FILE)
            ctypes.POINTER(glp_mpscp), # should be NULL
            ctypes.c_char_p,           # filename
        ]

        _lib.glp_write_mps.restype = ctypes.c_int
        _lib.glp_write_mps.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.c_int,              # format (GLP_MPS_DECK or GLP_MPS_FILE)
            ctypes.POINTER(glp_mpscp), # should be NULL
            ctypes.c_char_p,           # filename
        ]

        _lib.glp_write_lp.restype = ctypes.c_int
        _lib.glp_write_lp.argtypes = [
            ctypes.POINTER(glp_prob),
            ctypes.POINTER(glp_mpscp), # should be NULL
            ctypes.c_char_p,           # filename
        ]
        
        # LP Basis Factorization
        _lib.glp_get_bfcp.restype = None
        _lib.glp_get_bfcp.argtypes = [ctypes.POINTER(glp_prob), ctypes.POINTER(glp_bfcp)]

        _lib.glp_set_bfcp.restype = None
        _lib.glp_set_bfcp.argtypes = [ctypes.POINTER(glp_prob), ctypes.POINTER(glp_bfcp)]


        # Make accessible for front-end interface
        self._lib = _lib



'''GLPK integration.'''

import ctypes
from warnings import warn

import numpy as np
from scipy.sparse import coo_matrix
from scipy.optimize import OptimizeWarning, OptimizeResult
from scipy.optimize._linprog_util import _LPProblem, _clean_inputs

from ._glpk_defines import GLPK, glp_smcp, glp_iptcp, glp_bfcp, glp_iocp
from ._utils import _fill_prob


def glpk(
        c,
        A_ub=None,
        b_ub=None,
        A_eq=None,
        b_eq=None,
        bounds=None,
        solver='simplex',
        sense=GLPK.GLP_MIN,
        scale=True,
        maxit=GLPK.INT_MAX,
        timeout=GLPK.INT_MAX,
        basis_fac='luf+ft',
        message_level=GLPK.GLP_MSG_ERR,
        disp=False,
        simplex_options=None,
        ip_options=None,
        mip_options=None,
):
    '''GLPK ctypes interface.

    Parameters
    ----------
    c : 1-D array (n,)
        Array of objective coefficients.
    A_ub : 2-D array (m, n)
        scipy.sparse.coo_matrix
    b_ub : 1-D array (m,)
    A_eq : 2-D array (k, n)
        scipy.sparse.coo_matrix
    b_eq : 1-D array (k,)
    bounds : None or list (n,) of tuple (2,) or tuple (2,)
        The jth entry in the list corresponds to the jth objective coefficient.
        Each entry is made up of a tuple describing the bounds.
        Use None to indicate that there is no bound. By default, bounds are
        (0, None) (all decision variables are non-negative). If a single tuple
        (min, max) is provided, then min and max will serve as bounds for all
        decision variables.
    solver : { 'simplex', 'interior', 'mip' }
        Use simplex (LP/MIP) or interior point method (LP only).
        Default is ``simplex``.
    sense : { 'GLP_MIN', 'GLP_MAX' }
        Minimization or maximization problem.
        Default is ``GLP_MIN``.
    scale : bool
        Scale the problem. Default is ``True``.
    maxit : int
        Maximum number of iterations. Default is ``INT_MAX``.
    timout : int
        Limit solution time to ``timeout`` seconds.
        Default is ``INT_MAX``.
    basis_fac : { 'luf+ft', 'luf+cbg', 'luf+cgr', 'btf+cbg', 'btf+cgr' }
        LP basis factorization strategy. Default is ``luf+ft``.
        These are combinations of the following strategies:

            - ``luf`` : plain LU-factorization
            - ``btf`` : block triangular LU-factorization
            - ``ft`` : Forrest-Tomlin update
            - ``cbg`` : Schur complement + Bartels-Golub update
            - ``cgr`` : Schur complement + Givens rotation update

    message_level : { GLP_MSG_OFF, GLP_MSG_ERR, GLP_MSG_ON, GLP_MSG_ON, GLP_MSG_ALL, GLP_MSG_DBG }
        Verbosity level of logging to stdout.
        Only applied when ``disp=True``. Default is ``GLP_MSG_ERR``.
        One of the following:

            ``GLP_MSG_OFF`` : no output
            ``GLP_MSG_ERR`` : warning and error messages only
            ``GLP_MSG_ON`` : normal output
            ``GLP_MSG_ALL`` : full output
            ``GLP_MSG_DBG`` : debug output

    disp : bool
        Display output to stdout. Default is ``False``.
    simplex_options : dict
        Options specific to simplex solver. The dictionary consists of
        the following fields:

            - primal : { 'primal', 'dual', 'dualp' }
                Primal or two-phase dual simplex.
                Default is ``primal``. One of the following:

                    - ``primal`` : use two-phase primal simplex
                    - ``dual`` : use two-phase dual simplex
                    - ``dualp`` : use two-phase dual simplex, and if it fails,
                        switch to the primal simplex

            - init_basis : { 'std', 'adv', 'bib' }
                Choice of initial basis.  Default is 'adv'.
                One of the following:

                    - ``std`` : standard initial basis of all slacks
                    - ``adv`` : advanced initial basis
                    - ``bib`` : Bixby's initial basis

            - steep : bool
                Use steepest edge technique or standard "textbook"
                pricing.  Default is ``True`` (steepest edge).

            - ratio : { 'relax', 'norelax', 'flip' }
                Ratio test strategy. Default is ``relax``.
                One of the following:

                    - ``relax`` : Harris' two-pass ratio test
                    - ``norelax`` : standard "textbook" ratio test
                    - ``flip`` : long-step ratio test

            - tol_bnd : double
                Tolerance used to check if the basic solution is primal
                feasible. (Default: 1e-7).

            - tol_dj : double
                Tolerance used to check if the basic solution is dual
                feasible. (Default: 1e-7).

            - tol_piv : double
                Tolerance used to choose eligble pivotal elements of
                the simplex table. (Default: 1e-10).

            - obj_ll : double
                Lower limit of the objective function. If the objective
                function reaches this limit and continues decreasing,
                the solver terminates the search. Used in the dual simplex
                only. (Default: -DBL_MAX -- the largest finite float64).

            - obj_ul : double
                Upper limit of the objective function. If the objective
                function reaches this limit and continues increasing,
                the solver terminates the search. Used in the dual simplex
                only. (Default: +DBL_MAX -- the largest finite float64).

            - presolve : bool
                Use presolver (assumes ``scale=True`` and
                ``init_basis='adv'``. Default is ``True``.

            - exact : bool
                Use simplex method based on exact arithmetic.
                Default is ``False``. If ``True``, all other
                ``simplex_option`` fields are ignored.

    ip_options : dict
        Options specific to interior-pooint solver.
        The dictionary consists of the following fields:

            - ordering : { 'nord', 'qmd', 'amd', 'symamd' }
                Ordering algorithm used before Cholesky factorizaiton.
                Default is ``amd``. One of the following:

                    - ``nord`` : natural (original) ordering
                    - ``qmd`` : quotient minimum degree ordering
                    - ``amd`` : approximate minimum degree ordering
                    - ``symamd`` : approximate minimum degree ordering
                        algorithm for Cholesky factorization of symmetric
                        matrices.

    mip_options : dict
        Options specific to MIP solver.
        The dictionary consists of the following fields:

            - intcon : 1-D array
                Array of integer contraints, specified as the 0-based
                indices of the solution. Default is an empty array.
            - bincon : 1-D array
                Array of binary constraints, specified as the 0-based
                indices of the solution. If any indices are duplicated
                between ``bincon`` and ``intcon``, they will be
                considered as binary constraints. Default is an empty
                array.
            - nomip : bool
                consider all integer variables as continuous
                (allows solving MIP as pure LP). Default is ``False``.
            - branch : { 'first', 'last', 'mostf', 'drtom', 'pcost' }
                Branching rule. Default is ``drtom``.
                One of the following:

                    - ``first`` : branch on first integer variable
                    - ``last`` : branch on last integer variable
                    - ``mostf`` : branch on most fractional variable
                    - ``drtom`` : branch using heuristic by Driebeck and Tomlin
                    - ``pcost`` : branch using hybrid pseudocost heuristic
                                  (may be useful for hard instances)

            - backtrack : { 'dfs', 'bfs', 'bestp', 'bestb' }
                Backtracking rule. Default is ``bestb``.
                One of the following:

                    - ``dfs`` : backtrack using depth first search
                    - ``bfs`` : backtrack using breadth first search
                    - ``bestp`` : backtrack using the best projection heuristic
                    - ``bestb`` : backtrack using node with best local bound

            - preprocess : { 'none', 'root', 'all' }
                Preprocessing technique. Default is ``GLP_PP_ALL``.
                One of the following:

                    - ``none`` : disable preprocessing
                    - ``root`` : perform preprocessing only on the root level
                    - ``all`` : perform preprocessing on all levels

            - round : bool
                Simple rounding heuristic. Default is ``True``.

            - presolve : bool
                Use MIP presolver. Default is ``True``.

            - binarize : bool
                replace general integer variables by binary ones
                (only used if ``presolve=True``). Default is ``False``.

            - fpump : bool
                Apply feasibility pump heuristic. Default is ``False``.

            - proxy : int
                Apply proximity search heuristic (in seconds). Default is 60.

            - cuts : list of { 'gomory', 'mir', 'cover', 'clique', 'all' }
                Cuts to generate. Default is no cuts. List of the following:

                    - ``gomory`` : Gomory's mixed integer cuts
                    - ``mir`` : MIR (mixed integer rounding) cuts
                    - ``cover`` : mixed cover cuts
                    - ``clique`` : clique cuts
                    - ``all`` : generate all cuts above

            - tol_int : float
                Absolute tolerance used to check if optimal solution to the
                current LP relaxation is integer feasible.
                (Default: 1e-5).
            - tol_obj : float
                Relative tolerance used to check if the objective value in
                optimal solution to the current LP relaxation is not better
                than in the best known integer feasible solution.
                (Default: 1e-7).
            - mip_gap : float
                Relative mip gap tolerance. If the relative mip gap for
                currently known best integer feasiblesolution falls below
                this tolerance, the solver terminates the search. This allows
                obtaining suboptimal integer feasible solutions if solving the
                problem to optimality takes too long time.
                (Default: 0.0).
            - bound : float
                add inequality obj <= bound (minimization) or
                obj >= bound (maximization) to integer feasibility
                problem (assumes ``minisat=True``).

    Notes
    -----
    In general, don't change tolerances without a detailed understanding
    of their purposes.
    '''

    # Housekeeping
    if simplex_options is None:
        simplex_options = {}
    if ip_options is None:
        ip_options = {}
    if mip_options is None:
        mip_options = {}

    # Create and fill the GLPK problem struct
    prob, lp = _fill_prob(c, A_ub, b_ub, A_eq, b_eq, bounds, sense, 'problem-name')
    c, A_ub, b_ub, A_eq, b_eq, bounds, _x0 = lp

    # Get the library
    _lib = GLPK()._lib

    # Scale the problem
    no_need_explict_scale = (solver == "simplex" and 
                             simplex_options.get("presolve"))
    if not no_need_explict_scale and scale:
        _lib.glp_scale_prob(prob, GLPK.GLP_SF_AUTO) # do auto scaling for now


    # Select basis factorization method
    bfcp = glp_bfcp()
    _lib.glp_get_bfcp(prob, ctypes.byref(bfcp))
    bfcp.type = {
        'luf+ft': GLPK.GLP_BF_LUF + GLPK.GLP_BF_FT,
        'luf+cbg': GLPK.GLP_BF_LUF + GLPK.GLP_BF_BG,
        'luf+cgr': GLPK.GLP_BF_LUF + GLPK.GLP_BF_GR,
        'btf+cbg': GLPK.GLP_BF_BTF + GLPK.GLP_BF_BG,
        'btf+cgr': GLPK.GLP_BF_BTF + GLPK.GLP_BF_GR,
    }[basis_fac]
    _lib.glp_set_bfcp(prob, ctypes.byref(bfcp))

    # Run the solver
    if solver == 'simplex':

        # Construct an initial basis
        basis = simplex_options.get('init_basis', 'adv')
        basis_fun = {
            'std': _lib.glp_std_basis,
            'adv': _lib.glp_adv_basis,
            'bib': _lib.glp_cpx_basis,
        }[basis]
        basis_args = [prob]
        if basis == 'adv':
            # adv must have 0 as flags argument
            basis_args.append(0)
        basis_fun(*basis_args)

        # Make control structure
        smcp = glp_smcp()
        _lib.glp_init_smcp(ctypes.byref(smcp))

        # Set options
        smcp.msg_lev = message_level*disp
        smcp.meth = {
            'primal': GLPK.GLP_PRIMAL,
            'dual': GLPK.GLP_DUAL,
            'dualp': GLPK.GLP_DUALP,
        }[simplex_options.get('method', 'primal')]
        smcp.pricing = {
            True: GLPK.GLP_PT_PSE,
            False: GLPK.GLP_PT_STD,
        }[simplex_options.get('steep', True)]
        smcp.r_test = {
            'relax': GLPK.GLP_RT_HAR,
            'norelax': GLPK.GLP_RT_STD,
            'flip': GLPK.GLP_RT_FLIP,
        }[simplex_options.get('ratio', 'relax')]
        smcp.tol_bnd = simplex_options.get('tol_bnd', 1e-7)
        smcp.tol_dj = simplex_options.get('tol_dj', 1e-7)
        smcp.tol_piv = simplex_options.get('tol_piv', 1e-10)
        if simplex_options.get('obj_ll', False):
            smcp.obj_ll = simplex_options['obj_ll']
        if simplex_options.get('obj_ul', False):
            smcp.obj_ul = simplex_options['obj_ul']
        smcp.it_lim = maxit
        smcp.tm_lim = timeout
        smcp.presolve = {
            True: GLPK.GLP_ON,
            False: GLPK.GLP_OFF,
        }[simplex_options.get('presolve', True)]

        # Simplex driver
        if simplex_options.get('exact', False):
            ret_code = _lib.glp_exact(prob, ctypes.byref(smcp))
        else:
            ret_code = _lib.glp_simplex(prob, ctypes.byref(smcp))
        if ret_code != GLPK.SUCCESS:
            warn('GLPK simplex not successful!', OptimizeWarning)
            return OptimizeResult({
                'message': GLPK.RET_CODES[ret_code],
            })

        # Figure out what happened
        status = _lib.glp_get_status(prob)
        message = GLPK.STATUS_CODES[status]
        res = OptimizeResult({
            'status': status,
            'message': message,
            'success': status == GLPK.GLP_OPT,
        })

        # We can read a solution:
        if status == GLPK.GLP_OPT:

            res.fun = _lib.glp_get_obj_val(prob)
            res.x = np.array([_lib.glp_get_col_prim(prob, ii) for ii in range(1, _lib.glp_get_num_cols(prob)+1)])
            res.dual = np.array([_lib.glp_get_row_dual(prob, ii) for ii in range(1, _lib.glp_get_num_rows(prob)+1)])

            # We don't get slack without doing sensitivity analysis since GLPK
            # uses auxiliary variables instead of slack!
            res.slack = b_ub - A_ub @ res.x
            res.con = b_eq - A_eq @ res.x

            # We shouldn't be reading this field... But we will anyways
            res.nit = prob.contents.it_cnt

    elif solver == 'interior':

        # Make a control structure
        iptcp = glp_iptcp()
        _lib.glp_init_iptcp(ctypes.byref(iptcp))

        # Set options
        iptcp.msg_lev = message_level*disp
        iptcp.ord_alg = {
            'nord': GLPK.GLP_ORD_NONE,
            'qmd': GLPK.GLP_ORD_QMD,
            'amd': GLPK.GLP_ORD_AMD,
            'symamd': GLPK.GLP_ORD_SYMAMD,
        }[ip_options.get('ordering', 'amd')]

        # Run the solver
        ret_code = _lib.glp_interior(prob, ctypes.byref(iptcp))
        if ret_code != GLPK.SUCCESS:
            warn('GLPK interior-point not successful!', OptimizeWarning)
            return OptimizeResult({
                'message': GLPK.RET_CODES[ret_code],
            })

        # Figure out what happened
        status = _lib.glp_ipt_status(prob)
        message = GLPK.STATUS_CODES[status]
        res = OptimizeResult({
            'status': status,
            'message': message,
            'success': status == GLPK.GLP_OPT,
        })

        # We can read a solution:
        if status == GLPK.GLP_OPT:

            res.fun = _lib.glp_ipt_obj_val(prob)
            res.x = np.array([_lib.glp_ipt_col_prim(prob, ii) for ii in range(1, _lib.glp_get_num_cols(prob)+1)])
            res.dual = np.array([_lib.glp_ipt_row_dual(prob, ii) for ii in range(1, _lib.glp_get_num_rows(prob)+1)])

            # We don't get slack without doing sensitivity analysis since GLPK uses
            # auxiliary variables instead of slack!
            res.slack = b_ub - A_ub @ res.x
            res.con = b_eq - A_eq @ res.x

            # We shouldn't be reading this field... But we will anyways
            res.nit = prob.contents.it_cnt

    elif solver == 'mip':

        # Make a control structure
        iocp = glp_iocp()
        _lib.glp_init_iocp(ctypes.byref(iocp))

        # Make variables integer- and binary-valued
        if not mip_options.get('nomip', False):
            intcon = mip_options.get('intcon', [])
            for jj in intcon:
                _lib.glp_set_col_kind(prob, jj+1, GLPK.GLP_IV)
            bincon = mip_options.get('bincon', [])
            for jj in bincon:
                _lib.glp_set_col_kind(prob, jj+1, GLPK.GLP_BV)

        # Set options
        iocp.msg_lev = message_level*disp
        iocp.br_tech = {
            'first': GLPK.GLP_BR_FFV,
            'last': GLPK.GLP_BR_LFV,
            'mostf': GLPK.GLP_BR_MFV,
            'drtom': GLPK.GLP_BR_DTH,
            'pcost': GLPK.GLP_BR_PCH,
        }[mip_options.get('branch', 'drtom')]
        iocp.bt_tech = {
            'dfs': GLPK.GLP_BT_DFS,
            'bfs': GLPK.GLP_BT_BFS,
            'bestp': GLPK.GLP_BT_BPH,
            'bestb': GLPK.GLP_BT_BLB,
        }[mip_options.get('backtrack', 'bestb')]
        iocp.pp_teck = {
            'none': GLPK.GLP_PP_NONE,
            'root': GLPK.GLP_PP_ROOT,
            'all': GLPK.GLP_PP_ALL,
        }[mip_options.get('preprocess', 'all')]
        iocp.sr_heur = {
            True: GLPK.GLP_ON,
            False: GLPK.GLP_OFF,
        }[mip_options.get('round', True)]
        iocp.fp_heur = {
            True: GLPK.GLP_ON,
            False: GLPK.GLP_OFF,
        }[mip_options.get('fpump', False)]

        ps_tm_lim = mip_options.get('proxy', 60)
        if ps_tm_lim:
            iocp.ps_heur = GLPK.GLP_ON
            iocp.ps_tm_lim = ps_tm_lim*1000
        else:
            iocp.ps_heur = GLPK.GLP_OFF
            iocp.ps_tm_lim = 0

        cuts = set(list(mip_options.get('cuts', [])))
        if 'all' in cuts:
            cuts = {'gomory', 'mir', 'cover', 'clique'}
        if 'gomory' in cuts:
            iocp.gmi_cuts = GLPK.GLP_ON
        if 'mir' in cuts:
            iocp.mir_cuts = GLPK.GLP_ON
        if 'cover' in cuts:
            iocp.cov_cuts = GLPK.GLP_ON
        if 'clique' in cuts:
            iocp.clq_cuts = GLPK.GLP_ON

        iocp.tol_int = mip_options.get('tol_int', 1e-5)
        iocp.tol_obj = mip_options.get('tol_obj', 1e-7)
        iocp.mip_gap = mip_options.get('mip_gap', 0.0)
        iocp.tm_lim = timeout
        iocp.presolve = {
            True: GLPK.GLP_ON,
            False: GLPK.GLP_OFF,
        }[mip_options.get('presolve', True)]
        iocp.binarize = {
            True: GLPK.GLP_ON,
            False: GLPK.GLP_OFF,
        }[mip_options.get('binarize', False)]

        # Run the solver
        ret_code = _lib.glp_intopt(prob, ctypes.byref(iocp))
        if ret_code != GLPK.SUCCESS:
            warn('GLPK interior-point not successful!', OptimizeWarning)
            return OptimizeResult({
                'message': GLPK.RET_CODES[ret_code],
            })

        # Figure out what happened
        status = _lib.glp_mip_status(prob)
        message = GLPK.STATUS_CODES[status]
        res = OptimizeResult({
            'status': status,
            'message': message,
            'success': status in [GLPK.GLP_OPT, GLPK.GLP_FEAS],
        })

        # We can read a solution:
        if res.success:
            res.fun = _lib.glp_mip_obj_val(prob)
            res.x = np.array([_lib.glp_mip_col_val(prob, ii) for ii in range(1, len(c)+1)])

    else:
        raise ValueError('"%s" is not a recognized solver.' % solver)

    # We're done, cleanup!
    _lib.glp_delete_prob(prob)

    # Map status codes to scipy:
    # res.status = {
    #     GLPK.GLP_OPT: 0,
    # }[res.status]

    return res


if __name__ == '__main__':
    pass



import ctypes
import pathlib
import os

import numpy as np
from scipy.sparse import coo_matrix

from ._glpk_defines import GLPK
from ._utils import _fill_prob


def mpsread(filename, fmt=GLPK.GLP_MPS_FILE, ret_glp_prob=False, read_additional_info=False):
    '''Read an MPS file.

    Parameters
    ----------
    filename : str
        Path to MPS file.
    fmt : { GLP_MPS_DECK, GLP_MPS_FILE }
        Type of MPS file (original or free-format). Default is
        free-format (``GLP_MPS_FILE``).
    ret_glp_prob : bool
        Return the ``glp_prob`` structure. If ``False``, `linprog` style
        matrices and bounds are returned, i.e., ``A_ub``, ``b_ub``, etc.
        Default is ``False``.
    read_additional_info : bool
        Return additional info dictionary, which contains:
            col_names - column names
            row_names - row names
            row_types - row types (GLP_FR, GLP_LO, GLP_UP, GLP_DB, GLP_FX)
            col_types - column types (GLP_FR, GLP_LO, GLP_UP, GLP_DB, GLP_FX)
            prob_name - problem name
            obj_name  - objective function name
            obj_dir   - optimization direction flag (GLP_MIN, GLP_MAX)
        All types constants (e.g. GLP_FR, GLP_MIN)  are contained in GLPK.{constant name}
        If ``True``, additional info is returned.
        Default is ``False``.
    '''

    _lib = GLPK()._lib

    # Populate a glp_prob structure
    prob = _lib.glp_create_prob()
    filename = str(pathlib.Path(filename).expanduser().resolve())
    _lib.glp_read_mps(prob, fmt, None, filename.encode())

    # Return the GLPK object if requested
    if ret_glp_prob:
        return prob

    # Otherwise read the matrices
    n = _lib.glp_get_num_cols(prob)
    c = np.empty((n,))
    bounds = []
    for ii in range(1, n+1):
        c[ii-1] = _lib.glp_get_obj_coef(prob, ii)

        # get lb and ub; have to do it like this for some reason...
        tipe = _lib.glp_get_col_type(prob, ii)
        if tipe == GLPK.GLP_FR:
            bnd = (-np.inf, np.inf)
        elif tipe == GLPK.GLP_LO:
            bnd = (_lib.glp_get_col_lb(prob, ii), np.inf)
        elif tipe == GLPK.GLP_UP:
            bnd = (-np.inf, _lib.glp_get_col_ub(prob, ii))
        elif tipe == GLPK.GLP_DB:
            bnd = (_lib.glp_get_col_lb(prob, ii), _lib.glp_get_col_ub(prob, ii))
        else:
            bnd = (_lib.glp_get_col_lb(prob, ii), _lib.glp_get_col_lb(prob, ii))
        bounds.append(bnd)

    m = _lib.glp_get_num_rows(prob)
    nnz = _lib.glp_get_num_nz(prob)
    ub_cols, ub_rows, ub_vals = [], [], []
    eq_cols, eq_rows, eq_vals = [], [], []
    b_ub, b_eq = [], []
    indarr = ctypes.c_int*(n+1)
    valarr = ctypes.c_double*(n+1)
    col_ind = indarr()
    row_val = valarr()
    for ii in range(1, m+1):
        row_type = _lib.glp_get_row_type(prob, ii)
        l = _lib.glp_get_mat_row(prob, ii, col_ind, row_val)
        if l == 0:
            # if there are no elements in the row, skip,
            # otherwise we would add an extra element to RHS
            continue

        if row_type == GLPK.GLP_UP:
            b_ub.append(_lib.glp_get_row_ub(prob, ii))
            for jj in range(1, l+1):
                ub_cols.append(col_ind[jj] - 1)
                ub_rows.append(ii - 1)
                ub_vals.append(row_val[jj])
        elif row_type == GLPK.GLP_FX:
            b_eq.append(_lib.glp_get_row_lb(prob, ii))
            for jj in range(1, l+1):
                eq_cols.append(col_ind[jj] - 1)
                eq_rows.append(ii - 1)
                eq_vals.append(row_val[jj])
        elif row_type == GLPK.GLP_LO:
            b_ub.append(-1*_lib.glp_get_row_lb(prob, ii))
            for jj in range(1, l+1):
                ub_cols.append(col_ind[jj] - 1)
                ub_rows.append(ii - 1)
                ub_vals.append(-1*row_val[jj])
        else:
            raise NotImplementedError()

    if ub_vals:
        # Converting to csc_matrix gets rid of all-zero rows
        A_ub = coo_matrix((ub_vals, (ub_rows, ub_cols)), shape=(max(ub_rows)+1, n))
        A_ub = A_ub.tocsc()
        A_ub = A_ub[A_ub.getnnz(axis=1) > 0]
    else:
        A_ub = None
        b_ub = None

    if eq_vals:
        A_eq = coo_matrix((eq_vals, (eq_rows, eq_cols)), shape=(max(eq_rows)+1, n))
        A_eq = A_eq.tocsc()
        A_eq = A_eq[A_eq.getnnz(axis=1) > 0]
    else:
        A_eq = None
        b_eq = None

    if read_additional_info:
        additional_info = {'col_names': [_lib.glp_get_col_name(prob, ii).decode('utf-8') for ii in range(1, n + 1)],
                           'row_names': [_lib.glp_get_row_name(prob, ii).decode('utf-8') for ii in range(1, m + 1)],
                           'col_types': [_lib.glp_get_col_type(prob, ii) for ii in range(1, n + 1)],
                           'row_types': [_lib.glp_get_row_type(prob, ii) for ii in range(1, m + 1)],
                           'prob_name': _lib.glp_get_prob_name(prob), 'obj_name': _lib.glp_get_obj_name(prob),
                           'obj_dir': _lib.glp_get_obj_dir(prob)}

        return (c, A_ub, b_ub, A_eq, b_eq, bounds, additional_info)
    else:
        return(c, A_ub, b_ub, A_eq, b_eq, bounds)


def mpswrite(
        c,
        A_ub=None,
        b_ub=None,
        A_eq=None,
        b_eq=None,
        bounds=None,
        integrality=None,
        sense=GLPK.GLP_MIN,
        filename='prob.mps',
        fmt=GLPK.GLP_MPS_FILE):
    '''Write an MPS file.'''

    filename = pathlib.Path(filename)
    prob, _lp = _fill_prob(c, A_ub, b_ub, A_eq, b_eq, bounds, integrality, sense, filename.stem)

    # Get the library
    _lib = GLPK()._lib

    # Call write
    res = _lib.glp_write_mps(prob, fmt, None, str(filename).encode())
    return res == 0


def lpwrite(c,
        A_ub=None,
        b_ub=None,
        A_eq=None,
        b_eq=None,
        bounds=None,
        sense=GLPK.GLP_MIN,
        filename='prob.lp'):
    '''Write a CPLEX LP file.'''

    filename = pathlib.Path(filename)
    prob, _lp = _fill_prob(c, A_ub, b_ub, A_eq, b_eq, bounds, sense, filename.stem)

    # Get the library
    _lib = GLPK()._lib

    # Call write
    res = _lib.glp_write_lp(prob, None, str(filename).encode())
    return res == 0



'''Install scikit-glpk bindings.'''

import pathlib
from setuptools import find_packages
from distutils.core import Extension, setup
from distutils.command.build_ext import build_ext as _build_ext

GLPK_SRC_DIR = pathlib.Path('glpk-4.65/src')


def scrape_makefile_list(filename, START_TOKEN, END_TOKEN):
    '''Grab tags from GLPK makefile.'''
    with open(str(filename), 'r', encoding='utf-8') as f:
        _contents = f.read()
        sidx = _contents.find(START_TOKEN)
        eidx = _contents.find(END_TOKEN)
        lines = _contents[sidx+len(START_TOKEN):eidx].splitlines()
        return [str(_l.replace('\\', '').strip()) for _l in lines]


class build_ext(_build_ext):
    '''Override get_export_symbols to provide them for Windows DLL.'''
    def get_export_symbols(self, ext):
        '''Only for generating Windows DLL.'''
        def_file = GLPK_SRC_DIR / '../w64/glpk_4_65.def'
        return scrape_makefile_list(def_file, 'EXPORTS\n', ';; end of file ;;')


# Get sources for GLPK
makefile = GLPK_SRC_DIR / 'Makefile.am'
sources = scrape_makefile_list(
    makefile, 'libglpk_la_SOURCES = \\\n', '\n## eof ##')
sources = [str(GLPK_SRC_DIR / _s) for _s in sources]

# Get include dirs for GLPK
include_dirs = scrape_makefile_list(
    makefile, 'libglpk_la_CPPFLAGS = \\\n', '\nlibglpk_la_LDFLAGS')
include_dirs = [
    str(GLPK_SRC_DIR / _d[len('-I($srcdir)/'):]) for _d in include_dirs]


setup(
    name='scikit-glpk',
    version='0.4.4',
    author='Nicholas McKibben',
    author_email='nicholas.bgp@gmail.com',
    url='https://github.com/mckib2/scikit-glpk',
    license='MIT',
    description='Python linprog interface for GLPK',
    long_description=open('README.rst', encoding='utf-8').read(),
    packages=find_packages(),
    keywords='glpk linprog scikit',
    install_requires=open('requirements.txt', encoding='utf-8').read().split(),
    python_requires='>=3.5',

    ext_modules=[
        Extension(
            'glpk4_65',
            sources=sources,
            include_dirs=include_dirs,
            language='c',
        )
    ],
    cmdclass={'build_ext': build_ext},
)



'''Wrappers over utility API routines.'''

import ctypes

import numpy as np
from scipy.sparse import coo_matrix
from scipy.optimize._linprog_util import _LPProblem, _clean_inputs

from ._glpk_defines import GLPK


def _convert_bounds(processed_bounds):
    bounds = [None]*len(processed_bounds)
    for ii, (lb, ub) in enumerate(processed_bounds):
        if lb in {-np.inf, None} and ub in {np.inf, None}:
            # -inf < x < inf
            bounds[ii] = (GLPK.GLP_FR, 0, 0)
        elif lb in {-np.inf, None}:
            # -inf < x <= ub
            bounds[ii] = (GLPK.GLP_UP, 0, ub)
        elif ub in {np.inf, None}:
            # lb <= x < inf
            bounds[ii] = (GLPK.GLP_LO, lb, 0)
        elif lb < ub:
            # lb <= x <= ub
            bounds[ii] = (GLPK.GLP_DB, lb, ub)
        else:
            # lb == x == up
            bounds[ii] = (GLPK.GLP_FX, lb, ub)
    return bounds


def _fill_prob(c, A_ub, b_ub, A_eq, b_eq, bounds, integrality, sense, prob_name):
    '''Create and populate GLPK prob struct from linprog definition.'''

    # Housekeeping
    # TODO: modify to use integrality after https://github.com/mckib2/scipy/pull/28 makes it into scipy master
    lp = _clean_inputs(_LPProblem(c, A_ub, b_ub, A_eq, b_eq, bounds, None))
    c, A_ub, b_ub, A_eq, b_eq, processed_bounds, _x0 = lp

    # coo for (i, j, val) format
    A = coo_matrix(np.concatenate((A_ub, A_eq), axis=0))

    # Convert linprog-style bounds to GLPK-style bounds
    bounds = _convert_bounds(processed_bounds)

    # Get the library
    _lib = GLPK()._lib

    # Create problem instance
    prob = _lib.glp_create_prob()

    # Give problem a name
    _lib.glp_set_prob_name(prob, prob_name.encode())

    # Set objective name
    _lib.glp_set_obj_name(prob, b'obj-name')

    # Set objective sense
    _lib.glp_set_obj_dir(prob, sense)

    # Set objective coefficients and column bounds
    first_col = _lib.glp_add_cols(prob, len(c))
    for ii, (c0, bnd) in enumerate(zip(c, bounds)):
        _lib.glp_set_obj_coef(prob, ii + first_col, c0)
        _lib.glp_set_col_name(prob, ii + first_col, b'c%d' % ii) # name is c[idx], idx is 0-based index

        if bnd is not None:
            _lib.glp_set_col_bnds(prob, ii + first_col, bnd[0], bnd[1], bnd[2])
        # else: default is GLP_FX with lb=0, ub=0

        if integrality and integrality[ii] == 1:
            _lib.glp_set_col_kind(prob, ii + first_col, GLPK.GLP_IV)

    # Need to load both matrices at the same time
    first_row = _lib.glp_add_rows(prob, A.shape[0])

    # prepend an element and make 1-based index
    # b/c GLPK expects indices starting at 1
    nnz = A.nnz
    rows = np.concatenate(([-1], A.row + first_row)).astype(ctypes.c_int)
    cols = np.concatenate(([-1], A.col + first_col)).astype(ctypes.c_int)
    values = np.concatenate(([0], A.data)).astype(ctypes.c_double)
    _lib.glp_load_matrix(
        prob,
        nnz,
        rows,
        cols,
        values,
    )

    # Set row bounds
    # Upper bounds (b_ub):
    for ii, b0 in enumerate(b_ub):
        # lb is ignored for upper bounds
        _lib.glp_set_row_bnds(prob, ii + first_row, GLPK.GLP_UP, 0, b0)
    # Equalities (b_eq)
    for ii, b0 in enumerate(b_eq):
        _lib.glp_set_row_bnds(prob, ii + first_row + len(b_ub), GLPK.GLP_FX, b0, b0)

    return prob, lp
